# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-02-27
#


import miaou


miaou.generate(
    site_url="http://localhost:21761/zentao",
    # dev_url="http://localhost:21820/dev-api-index.html",
    username="admin",
    password="Lton2008@",
    combined_print=True,
    output_path=".",
    scanner="selenium",
    # scanner="api",
)


# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-02-27
#


__version__ = "0.2.1"


from miaou.app import Application
from miaou.scanner.base import Scanner


def generate(**attrs):
    """main caller"""

    Application(**attrs).main()


# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-03-01
#


class Scanner:
    """scanner base class"""

    def open(self, site_url, username, password):
        """tear up works"""

        pass

    def close(self):
        """tear down works"""

        pass

    def get_module_groups(self, dev_url):
        """get module group page urls

        return url string list [url]
        """

        pass

    def get_apis(self, api_url):
        """get apis from api page

        return list like [{name, method, path, params}]
        """

        pass

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-03-02
#


import re
import json
import urllib
import requests
import miaou.logger as logger
from miaou.scanner.base import Scanner


class ApiScanner(Scanner):
    """scan specs by zentao apis"""

    def __init__(self, config):
        """initializer"""

        pass

    def open(
        self,
        site_url,
        username,
        password,
    ):
        """tear up works"""

        session_url = urllib.parse.urljoin(
            site_url, "api-getSessionID.json")
        login_url = urllib.parse.urljoin(
            site_url, "user-login.json")

        return self._get_session(session_url) and \
            self._login(login_url, username, password)

    def get_module_groups(self, dev_url):
        """get module group page urls

        return url string list [url]
        """

        urls = []

        # change html to json
        dev_json_url = re.sub(r"\.html.*", r".json", dev_url)
        api_url_prefix = re.sub(r"\.html.*", "", dev_url)

        response = requests.get(
            dev_json_url,
            params={self.session_name: self.session_id}
        ).json()

        if "success" == response.get("status"):
            data = json.loads(response.get("data", r"{}"))

            # get modules, it's a dict like
            #   {module_name: {api_name: api_name}}
            modules = data.get("modules", {})

            # get apis
            for module, apis in modules.items():
                if isinstance(apis, dict):
                    _apis = apis.keys()
                elif isinstance(apis, list):
                    _apis = apis
                else:
                    _apis = []
                    logger.warning(
                        "unknown apis data type of module '%s'" % module)

                for api_name in _apis:
                    api = "%s-%s.json" % (api_url_prefix, api_name)
                    urls.append(api)
                    logger.info("got '%s'" % api)

        else:
            logger.error("fail to get module groups")

        return urls

    def get_apis(self, api_url):
        """get apis from api page

        return list like [{name, method, path, params}]
        """

        apis = []

        response = requests.get(
            api_url,
            params={self.session_name: self.session_id}
        ).json()

        if "success" == response.get("status"):
            data = json.loads(response.get("data", r"{}"))

            # selected module, it's api prefix
            module = data.get("selectedModule", "")

            # get apis, it's a list like
            #   [{name, post, param, desc...}]
            responded_apis = data.get("apis", [])
            for responded_api in responded_apis:
                api = {}

                # api name
                name = responded_api.get("name", "")

                api["name"] = "%s_%s" % (module, name)
                logger.info("got name: %s" % api["name"])

                # api path
                api["path"] = "%s-%s" % (module, name)
                logger.info("- got path: %s" % api["path"])

                # api method
                if responded_api.get("post", False):
                    api["method"] = "POST"
                else:
                    api["method"] = "GET"

                # api params
                params = responded_api.get("param", [])
                if isinstance(params, dict):
                    api["params"] = list(params.keys())
                    logger.info("- got params: %s" % api["params"])
                else:
                    api["params"] = []

                apis.append(api)
        else:
            logger.error("fail to get api from '%s'" % api_url)

        return apis

# protected
    def _get_session(self, session_url):
        """get zentao session name and session id"""

        response = requests.get(session_url).json()

        if "success" == response.get("status"):
            # get data list and return to caller
            session = json.loads(response["data"])
            self.session_name = session["sessionName"]
            self.session_id = session["sessionID"]

            logger.info("got session")
            return True
        else:
            # fail to get session
            logger.error("fail to get session")
            return False

    def _login(self, login_url, username, password):
        """login zentao with username and password"""

        response = requests.post(
            login_url,
            params={
                "account": username,
                "password": password,
                self.session_name: self.session_id
            }
        ).json()

        if "success" == response.get("status"):
            logger.info("succeed to login zentao")
            return True
        else:
            # fail to sign in
            logger.error("fail to login zentao")
            return False

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-02-26
#


import os
import time
import miaou.logger as logger
import miaou.utils as utils


class Specs:
    """api specs"""

    def __init__(self):
        """initializer"""

        self.apis = {}

# public
    def append(self, module, apis):
        """append api spec"""

        if module not in self.apis:
            self.apis[module] = []

        self.apis[module] += apis

    def print(self, output_path=".", combined=True):
        """print to yaml file"""

        if not os.path.exists(output_path):
            logger.error("the output path does not exsit.")
            return

        # make result folder
        result_path = os.path.join(
            output_path,
            "pyzentao_specs_%s" % time.strftime("%Y_%m_%d_%H_%M_%S")
        )

        os.makedirs(result_path)

        # print to file
        if combined:
            # print specs to one file
            self._print_to_file(
                path=os.path.join(result_path, "specs.yml"),
                specs=utils.flatten(self.apis.values())
            )
        else:
            # print specs by module
            for key, apis in self.apis.items():
                self._print_to_file(
                    path=os.path.join(
                        result_path, self._url_to_filename(key)),
                    specs=apis
                )

# private
    def _print_to_file(self, path, specs):
        """print spec to yaml file"""

        text = ""
        for spec in specs:
            text += "%s:\n" % spec["name"]
            text += "    method: %s\n" % spec["method"]
            text += "    path: %s\n" % spec["path"]
            if len(spec["params"]) > 0:
                text += "    params:\n"
                for param in spec["params"]:
                    text += "        - %s\n" % param
            text += "\n"

        logger.info("print to '%s'" % os.path.basename(path))
        with open(path, mode="w", encoding="utf-8") as f:
            f.write(text)

    @classmethod
    def _url_to_filename(self, url):
        """convert url to spec file name"""

        filename = url.split("/")[-1]
        filename = filename.split("?")[0]
        filename = filename.split("-")[-1]
        filename = filename.split(".")[0]
        return "%s.yml" % filename

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-02-26
#


import click


def info(message):
    """print normal message in stdout."""

    click.echo(message)


def stage(message):
    """print stage message in stdout."""

    click.secho(message, fg="green")


def warning(message):
    """print warning message in stdout."""

    click.secho("[WARNING] " + message, fg="yellow")


def error(message):
    """print error message in stdout."""

    click.secho("[ERROR] " + message, fg="red")


# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-02-27
#


from miaou.scanner.selenium_scanner import SeleniumScanner
from miaou.scanner.api_scanner import ApiScanner


class ScannerFactory:
    """scanner factory"""

    _map = {
        "selenium": SeleniumScanner,
        "api": ApiScanner,
    }

    @classmethod
    def get(cls, scanner, config={}):
        """get scanner by scanner type"""

        if isinstance(scanner, str):
            return cls._map.get(scanner, SeleniumScanner)(config)
        else:
            # maybe customized scanner
            return scanner

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-02-27
#


def flatten(l):
    """make list flatten"""

    for _ in l:
        if isinstance(_, list):
            yield from flatten(_)
        else:
            yield _


# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-02-26
#


import urllib
import miaou.logger as logger
from miaou.scanner import ScannerFactory
from miaou.specs import Specs


class Application:
    """main application"""

    def __init__(self, **attrs):
        """initializer, supported config attrs:

            - site_url: zentao site url, required
            - username: admin user name, required
            - password: admin password, required
            - dev_url: dev page url, auto generated by site url
            - scanner: scanner type, currently supported 'selenium', or
                    customized scanner instance
            - combined_print: print to one yaml or not, default is True
            - output_path: output path, default is current directory
        """

        self.site_url = attrs.get("site_url", "")
        # check out url
        if not self.site_url.endswith("/"):
            self.site_url += "/"

        self.username = attrs.get("username", "admin")
        self.password = attrs.get("password", "")

        self.dev_url = attrs.get("dev_url", None)
        if self.dev_url is None:
            self.dev_url = urllib.parse.urljoin(
                self.site_url, "dev-api.html")

        self.scanner = ScannerFactory.get(
            attrs.get("scanner", "selenium"),
            attrs.get("config", {})
        )

        self.specs = Specs()
        self.combined_print = attrs.get("combined_print", True)
        self.output_path = attrs.get("output_path", ".")

# public
    def main(self):
        """main routine"""

        # login
        self.scanner.open(
            site_url=self.site_url,
            username=self.username,
            password=self.password
        )

        # get module groups
        logger.stage("get module groups from '%s'..." % self.dev_url)
        modules = self.scanner.get_module_groups(self.dev_url)

        # get apis
        logger.stage("get apis...")
        for module_url in modules:
            self.specs.append(
                module=module_url,
                apis=self.scanner.get_apis(module_url)
            )

        # print out
        logger.stage("print specs...")
        self.specs.print(
            output_path=self.output_path,
            combined=self.combined_print,
        )

        self.scanner.close()

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2023-02-27
#


import re
import time
import miaou.logger as logger
from miaou.scanner.base import Scanner
from selenium import webdriver
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.common.by import By


_step_interval = 2


class SeleniumScanner(Scanner):
    """selenium scanner"""

    def __init__(self, config):
        """initializer"""

        if "firefox" == config.get("driver", "chrome"):
            self.driver = webdriver.Firefox()
        else:
            self.driver = webdriver.Chrome()

        wait_timeout = config.get("wait_timeout", 3)
        self.driver.implicitly_wait(wait_timeout)
        self.wait = WebDriverWait(self.driver, wait_timeout)

    def open(self, site_url, username, password):
        """login zentao"""

        logger.stage("login zentao '%s'" % site_url)

        self.driver.get(site_url)
        time.sleep(_step_interval)

        account = self._wait_for_element("input#account")
        account.send_keys(username)

        passwd = self._wait_for_element("input[name=password]")
        passwd.send_keys(password)

        auth = self._wait_for_element("button#submit")
        # auth.send_keys(Keys.ENTER)
        auth.click()

        time.sleep(_step_interval)

    def close(self):
        """close scanner"""

        self.driver.close()

    def get_module_groups(self, dev_url):
        """get module group page urls

        return url string list
        """

        urls = []

        self.driver.get(dev_url)
        time.sleep(_step_interval)

        # check out iframe
        self._switch_to_content_iframe()

        # find module group urls
        mg_elements = self._wait_for_elements(
            "#mainContent #sidebar a")

        for element in mg_elements:
            url = element.get_attribute("href")
            logger.info("got '%s'" % url)
            urls.append(url)

        return urls

    def get_apis(self, api_url):
        """get apis from api page

        return list like [{name, method, path, params}]
        """

        apis = []

        logger.info("get api from '%s'..." % api_url)
        self.driver.get(api_url)
        time.sleep(_step_interval)

        # check out iframe
        self._switch_to_content_iframe()

        # find urls
        detail_elements = self._wait_for_elements(
            "#mainContent .main-col.main-content .detail")
        for element in detail_elements:
            url = element.find_element(By.CSS_SELECTOR, ".detail-title").text

            method, path, params = self._parse_url_to_api_items(url)
            name = path.replace("-", "_")

            apis.append({
                "name": name,
                "method": method,
                "path": path,
                "params": params
            })

        return apis

# private
    def _switch_to_content_iframe(self):
        """check out the iframe exists and switch into it"""

        try:
            logger.info("check out iframe...")
            ifr = self._wait_for_element(
                "iframe#appIframe-admin, iframe#appIframe-my")
            self.driver.switch_to.frame(ifr)
        except Exception:
            pass

    def _wait_for_element(
        self,
        selector,
        scroll_to=True,
        by=By.CSS_SELECTOR
    ):
        """wait for element, scroll window to the element position
        if scroll_to is True, make the element visible.
        """

        try:
            _element = self.wait.until(
                lambda driver: driver.find_element(by, selector))

            if scroll_to:
                self._scroll_to_element(_element)
        except Exception:
            logger.warning("element not found %s" % selector)
            _element = None
        return _element

    def _wait_for_elements(
        self,
        selector,
        by=By.CSS_SELECTOR
    ):
        """wait for elements by selector type"""

        return self.wait.until(
            lambda driver: driver.find_elements(by, selector))

    def _element_exists(
        self,
        selector,
        by=By.CSS_SELECTOR
    ):
        """check out the element existence"""

        try:
            _element = self.wait.until(
                lambda driver: driver.find_element(by, selector))
        except Exception:
            _element = None

        if _element is not None:
            logger.info("element found %s" % selector)
            return True
        else:
            logger.info("element not exists %s" % selector)
            return False

    def _scroll_to_element(self, element):
        """scroll window to the element position"""

        self.driver.execute_script(
            "arguments[0].scrollIntoView();",
            element
        )

    def _parse_url_to_api_items(self, url):
        """parse url to api items,
        return tuple contains (method, path, params)"""

        logger.info("parsing '%s'" % url)

        path = []
        params = []

        # remove namespace and suffix
        slices = re.split("\s+", url)
        method = slices[0].strip().upper()
        if "GET/POST" == method:
            method = "POST"

        url = slices[1].strip()
        for pattern in (
            r"^/zentao/",
            r"\.json.*$",
            r"^/",
        ):
            url = re.sub(pattern, "", url)

        # split path
        for t in url.split("-"):
            t = t.strip()
            if re.match("\[\w+\]", t) is None:
                logger.info("- got path '%s'" % t)
                path.append(t)
            else:
                t = t.replace(r"[", "")
                t = t.replace(r"]", "")
                logger.info("- got param '%s'" % t)
                params.append(t)

        return (method, "-".join(path), params)
# end
